/*
 * build: cc -o client client.c -lrdmacm
 * usage: client <servername> <val1> <val2>
 *
 * connects to server, sends val1 via RDMA write
 * and val2 via send
 * and receives val1 + val2 back from the server
 */

#include <jni.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <arpa/inet.h>

#include <infiniband/verbs.h>
#include <infiniband/arch.h>
#include <rdma/rdma_cma.h>

#include "RdmaOpsClient.h"

enum {
  RESOLVE_TIMEOUT_MS = 5000,
};

struct pdata {
  uint64_t 			buf_va;
  uint32_t 			buf_rkey;
};

  struct rdma_cm_id		*cm_id;
JNIEXPORT jint JNICALL Java_RdmaOpsClient_write(JNIEnv *env, jobject jobj,
                                         jstring argv1, jstring argv2,
                                         jstring argv3) {
  struct pdata 			server_pdata;

  struct rdma_event_channel 	*cm_channel;
  //struct rdma_cm_id		*cm_id;
  struct rdma_cm_event		*event;
  struct rdma_conn_param	conn_param = {};

  struct ibv_pd			*pd = NULL;
  struct ibv_comp_channel	*comp_chan;
  struct ibv_cq			*cq;
  struct ibv_cq			*evt_cq;
  struct ibv_mr			*mr;
  struct ibv_qp_init_attr	qp_attr = {};
  struct ibv_sge		sge;
  struct ibv_send_wr		send_wr = {};
  struct ibv_send_wr		*bad_send_wr;
  struct ibv_recv_wr		recv_wr = {};
  struct ibv_recv_wr		*bad_recv_wr;
  struct ibv_wc			wc;
  void				*cq_context;

  struct addrinfo		*res, *t;
  struct addrinfo		hints = {
	.ai_family = AF_INET,
	.ai_socktype = SOCK_STREAM
  };

  int					n;
  uint32_t				*buf;
  jint  				err;
  //jboolean				iscopy = 0;
  char*					message;

  /* set up RDMA CM structures */
  
  /* DESC: rdma_create_event_channel creates a rdma event channel
   * on rdma channel adapter
   */
  cm_channel = rdma_create_event_channel();
  if(!cm_channel) {
	err = 1;
	return err;
  }

  /* DESC: rdma_create_id allocates rdma_cm_id, which is similar to 
   * a socket
   * ARGS: cm_channel is the event channel to receive events
   *       RDMA_PS_TCP is the protocol 
  */
  err = rdma_create_id(cm_channel, &cm_id, NULL, RDMA_PS_TCP);
  if(err)
	return err;

  /* convert 16 bit jstring into 8 bit characters */
  const char *arg0 = (*env)->GetStringUTFChars(env, argv1, 0);
  const char *arg1 = (*env)->GetStringUTFChars(env, argv1, 0);
  const char *arg2 = (*env)->GetStringUTFChars(env, argv2, 0);
  const char *arg3 = (*env)->GetStringUTFChars(env, argv3, 0);

  printf("arguments: %s - %s - %s - %s\n", arg0, arg1, arg2, arg3);

  n = getaddrinfo(arg1, "20079", &hints, &res);

  printf("n - %d\n", n);

  if(n < 0) {
	printf("inside\n");
	err = 1;
	return err;
  }

  printf("1\n");

  /* resolve server address and route */
  for(t = res; t; t = t->ai_next) {
	/* DESC: rdma_resolve_addr resolves remote address based on
	 * cm_id
	 * ARGS: cm_id is analogous to a socket
	 *	 t->ai_addr provides the address
	 *	 RESOLVE_TIMEOUT_MS is the timeout for the address
	 *		resolution request 
	 */
	err = rdma_resolve_addr(cm_id, NULL, t->ai_addr,
		RESOLVE_TIMEOUT_MS);

	if(!err)
		break;
  }
  if(err)
	return err;

  printf("2\n");

  err = rdma_get_cm_event(cm_channel, &event);
  printf("err - %d\n", err);
  if(err == 1)
	return err;

  printf("3\n");

  if(event->event != RDMA_CM_EVENT_ADDR_RESOLVED) {
	err = 1;
	return err;
  }

  printf("4\n");

  rdma_ack_cm_event(event);

  err = rdma_resolve_route(cm_id, RESOLVE_TIMEOUT_MS);
  if(err)
	return err;

  err = rdma_get_cm_event(cm_channel, &event);
  if(err)
	return err;

  if(event->event != RDMA_CM_EVENT_ROUTE_RESOLVED) {
	err = 1;
	return err;
  }

  printf("5\n");

  rdma_ack_cm_event(event);

  printf("6\n");fflush(stdout);
  printf("cm_id->port: %d\n", cm_id->port_num);fflush(stdout);
  printf("cm_id->verbs->cmd_fd: %d\n", cm_id->verbs->cmd_fd);fflush(stdout);

  /* create verbs objects now that we know which device to use */
  pd = ibv_alloc_pd(cm_id->verbs);
  printf("7\n");fflush(stdout);
  if(!pd) {
	err = 1;
	return err;
  }

  printf("8\n");

  comp_chan = ibv_create_comp_channel(cm_id->verbs);
  if(!comp_chan) {
	err = 1;
	return err;
  }

  cq = ibv_create_cq(cm_id->verbs, 2, NULL, comp_chan, 0);
  if(!cq) {
	err = 1;
	return err;
  }

  if(ibv_req_notify_cq(cq,0)) {
	err = 1;
	return err;
  }

  buf = calloc(2, sizeof(uint32_t));
  if(!buf) {
	err = 1;
	return err;
  }

  mr = ibv_reg_mr(pd, buf, 2 * sizeof(uint32_t), IBV_ACCESS_LOCAL_WRITE);
  if(!mr) {
	err = 1;
	return err;
  }

  qp_attr.cap.max_send_wr = 2;
  qp_attr.cap.max_send_sge = 1;
  qp_attr.cap.max_recv_wr = 1;
  qp_attr.cap.max_recv_sge = 1;

  qp_attr.send_cq = cq;
  qp_attr.recv_cq = cq;

  qp_attr.qp_type = IBV_QPT_RC;

  err = rdma_create_qp(cm_id, pd, &qp_attr);
  if(err)
	return err;

  conn_param.initiator_depth = 1;
  conn_param.retry_count = 7;

  message = "connecting to the server ...";
  printf("%s\n", message);

  /* connect to server */
  err = rdma_connect(cm_id, &conn_param);
  if(err)
	return err;

  err = rdma_get_cm_event(cm_channel, &event);
  if(err)
	return err;

//  if(event->event != RDMA_CM_EVENT_ESTABLISHED)
//	return 1;

  memcpy(&server_pdata, event->param.conn.private_data,
	sizeof server_pdata);

  rdma_ack_cm_event(event);

  /* prepost receive */
  sge.addr = (uintptr_t) buf;
  sge.length = sizeof(uint32_t);
  sge.lkey = mr->lkey;

  recv_wr.wr_id = 0;
  recv_wr.sg_list = &sge;
  recv_wr.num_sge = 1;

  if(ibv_post_recv(cm_id->qp, &recv_wr, &bad_recv_wr)) {
	err = 1;
	return 1;
  }

  message = "write/read data from server: ";
  printf("%s", message);

  /* write two integers to be added */
  buf[0] = strtoul(arg2, NULL, 0);
  buf[1] = strtoul(arg3, NULL, 0);

  printf("%d + %d = ", buf[0], buf[1]);

  /* htonl converts the byte order from host to network,
  this is to transfer data onto the network. */
  buf[0] = htonl(buf[0]);
  buf[1] = htonl(buf[1]);

  sge.addr = (uintptr_t) buf;
  sge.length = sizeof(uint32_t);
  sge.lkey = mr->lkey;

  send_wr.wr_id = 1;
  send_wr.opcode = IBV_WR_RDMA_WRITE;
  send_wr.sg_list = &sge;
  send_wr.num_sge = 1;
  send_wr.wr.rdma.rkey = ntohl(server_pdata.buf_rkey);
  send_wr.wr.rdma.remote_addr = ntohll(server_pdata.buf_va);

  if(ibv_post_send(cm_id->qp, &send_wr, &bad_send_wr)) {
	err = 1;
	return err;
  }

  sge.addr = (uintptr_t)buf + sizeof(uint32_t);
  sge.length = sizeof(uint32_t);
  sge.lkey = mr->lkey;

  send_wr.wr_id = 2;
  send_wr.opcode = IBV_WR_SEND;
  send_wr.send_flags = IBV_SEND_SIGNALED;
  send_wr.sg_list = &sge;
  send_wr.num_sge = 1;

  if(ibv_post_send(cm_id->qp, &send_wr, &bad_send_wr)) {
	err = 1;
	return err;
  }

  /* wait for receive completion */
  while(1) {
	if(ibv_get_cq_event(comp_chan, &evt_cq, &cq_context)) {
		err = 1;
		return err;
	}

	if(ibv_req_notify_cq(cq, 0)) {
		err = 1;
		return err;
	}

	if(ibv_poll_cq(cq, 1, &wc) != 1) {
		err = 1;
		return err;
	}

	if(wc.status != IBV_WC_SUCCESS) {
		err = 1;
		return err;
	}

	if(wc.wr_id == 0) {
		printf("%d\n", ntohl(buf[0]));
		err = 0;
		return err;
	}
  }

  /* deallocate the protection domain (pd) and with
   * it all the consumed resources.
   */
  err = ibv_dealloc_pd(pd);
  if(err)
	return err; 

  /* release the occupied native strings */
  (*env)->ReleaseStringUTFChars(env, argv1, arg0);
  (*env)->ReleaseStringUTFChars(env, argv1, arg1);
  (*env)->ReleaseStringUTFChars(env, argv2, arg2);
  (*env)->ReleaseStringUTFChars(env, argv3, arg3);

  err = 0;
  return err;
}


